#
# Copyright 2019-2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Auxiliary script to perform binary image arithmetic.

The computation is performed casting to np.float32.
"""

import argparse
from rich import print
from rich_argparse import RichHelpFormatter
import sys

from astropy.io import fits
import numpy as np
import pathlib

from numina.array.display.ximshow import ximshow_file, ximshow


def extnum_from_extname(file, extname):
    """Get the extension number from the extension name in a FITS file.

    Parameters
    ----------
    file : file object
        FITS file.
    extname : str
        Extension name.

    Returns
    -------
    int
        Extension number.
    """
    with fits.open(file) as hdulist:
        for i, hdu in enumerate(hdulist):
            if hdu.name.lower() == extname.lower():
                return i

    raise ValueError(f"Extension '{extname}' not found in {file.name}.")


def compute_operation(
    file1,
    file2,
    operation,
    extname1,
    extname2,
    output,
    overwrite=True,
    dtype='float32',
    display='none',
    args_z1z2=None,
    args_bbox=None,
    args_keystitle=None,
    args_geometry=None
):
    """Compute output = file1 operation file2.

    Parameters
    ----------
    file1 : file object
        First FITS file.
    file2 : file object or float
        Second FITS file or float number.
    operation : string
        Mathematical operation.
    extname1 : str
        Extension name of the first FITS file.
    extname2 : str
        Extension name of the second FITS file.
    output : file object
        Output FITS file.
    overwrite : bool
        If True, the output file can be overwritten.
    dtype : str
        Data type of the output image. Default is 'float32'.
    display : string
        Character string indication whether the images are displayed.
        Valid values are 'all', 'result' and 'none' (default).
    args_z1z2 : string or None
        String providing the image cuts tuple: z1, z2, minmax or None.
    args_bbox : string or None
        String providing the bounding box tuple: nc1, nc2, ns1, ns2.
    args_keystitle : string or None
        Tuple of FITS keywords.format: key1,key2,...,keyn.format
    args_geometry : string or None
        Tuple x,y,dx,dy to define the Qt backend geometry.

    """

    # read first FITS file
    with fits.open(file1) as hdulist:
        image_header1 = hdulist[extname1].header
        image1 = hdulist[extname1].data.astype(dtype)
    naxis = image_header1['naxis']
    naxis1 = image_header1['naxis1']
    naxis2 = image_header1['naxis2']

    # if required, display file1
    if display == 'all':
        ximshow_file(file1,
                     extnum=extnum_from_extname(file1, extname1)+1,  # extnum is 1-based in ximshow_file
                     args_z1z2=args_z1z2, args_bbox=args_bbox,
                     args_keystitle=args_keystitle,
                     args_geometry=args_geometry,
                     debugplot=12)

    # read second FITS file or number
    file2_is_a_file = True
    try:
        with fits.open(file2) as hdulist:
            image_header2 = hdulist[extname2].header
            image2 = hdulist[extname2].data.astype(dtype)
            naxis_ = image_header2['naxis']
            naxis1_ = image_header2['naxis1']
            naxis2_ = image_header2['naxis2']
            filename = file2
    except FileNotFoundError:
        file2_is_a_file = False
        image2 = np.zeros((naxis2, naxis1), dtype=dtype)
        if 'int' in dtype:
            image2 += int(file2)
        elif 'float' in dtype:
            image2 += float(file2)
        else:
            raise ValueError(f"Unsupported dtype: {dtype}. Use an integer or float type.")
        naxis_ = naxis
        naxis1_ = naxis1
        naxis2_ = naxis2
        filename = 'constant=' + str(file2)

    # if required, display file2
    if display == 'all':
        if file2_is_a_file:
            ximshow_file(file2,
                         extnum=extnum_from_extname(file2, extname2)+1,  # extnum is 1-based in ximshow_file
                         args_z1z2=args_z1z2, args_bbox=args_bbox,
                         args_keystitle=args_keystitle,
                         args_geometry=args_geometry,
                         debugplot=12)
        else:
            ximshow(image2,
                    title=filename,
                    z1z2=args_z1z2, image_bbox=args_bbox,
                    geometry=args_geometry,
                    debugplot=12)

    # second image is a single image row
    if naxis1 == naxis1_ and naxis2 > naxis2_:
        if naxis2_ == 1:
            image2 = np.tile(image2, naxis2).reshape((naxis2, naxis1))
            naxis2_ = naxis2

    # second image is a single image column
    if naxis2 == naxis2_ and naxis1 > naxis1_:
        if naxis1_ == 1:
            image2 = np.repeat(image2, naxis1).reshape((naxis2, naxis1))
            naxis1_ = naxis1

    # additional dimension checks
    if naxis != naxis_:
        raise ValueError("NAXIS values are different.")
    if naxis1 != naxis1_:
        raise ValueError("NAXIS1 values are different.")
    if naxis2 != naxis2_:
        raise ValueError("NAXIS2 values are different.")

    # compute operation
    if operation == "+":
        solution = image1 + image2
    elif operation == "-":
        solution = image1 - image2
    elif operation == "x":
        solution = image1 * image2
    elif operation == "/":
        solution = image1 / image2
    elif operation == "=":
        solution = image2.copy()
    else:
        raise ValueError("Unexpected operation=" + str(operation))

    # save output file
    hdu = fits.PrimaryHDU(solution.astype(dtype), image_header1)
    hdu.writeto(output, overwrite=overwrite)

    # if required, display result
    if display in ['all', 'result']:
        ximshow_file(output,
                     args_z1z2=args_z1z2, args_bbox=args_bbox,
                     args_keystitle=args_keystitle,
                     args_geometry=args_geometry,
                     debugplot=12)


def main(args=None):

    # parse command-line options
    parser = argparse.ArgumentParser(
        description="description: binary image arithmetic",
        formatter_class=RichHelpFormatter
    )
    # positional parameters
    parser.add_argument("file1",
                        help="First FITS image",
                        type=argparse.FileType('rb'))
    parser.add_argument("operation",
                        help="Arithmetic operation",
                        type=str,
                        choices=['+', '-', 'x', '/', '='])
    parser.add_argument("file2",
                        help="Second FITS image or number",
                        type=str)
    parser.add_argument("output",
                        help="Output FITS image",
                        type=str)
    # optional arguments
    parser.add_argument("--extname1", type=str,
                        help="Extension name of the first FITS file (default: 'PRIMARY').",
                        default='PRIMARY')
    parser.add_argument("--extname2", type=str,
                        help="Extension name of the second FITS file (default: 'PRIMARY').",
                        default='PRIMARY')
    parser.add_argument("--overwrite",
                        help="Overwrite output file if already exists",
                        action="store_true")
    parser.add_argument("--dtype",
                        help="Data type of the output image (default: float32)",
                        type=str,
                        choices=['uint8', 'int8', 'uint16', 'int16', 'uint32', 'int32', 'uint64', 'int64',
                                 'float32', 'float64'],
                        default='float32')
    parser.add_argument("--display",
                        help="Display images: all, result, none (default)",
                        default="none",
                        type=str,
                        choices=['all', 'result', 'none'])
    parser.add_argument("--z1z2",
                        help="tuple z1,z2, minmax or None (use zscale)")
    parser.add_argument("--bbox",
                        help="bounding box tuple: nc1,nc2,ns1,ns2")
    parser.add_argument("--keystitle",
                        help="tuple of FITS keywords.format: " +
                             "key1,key2,...keyn.'format'")
    parser.add_argument("--geometry",
                        help="Tuple x,y,dx,dy indicating window geometry",
                        default="0,0,640,480")
    parser.add_argument("--echo",
                        help="Display full command line",
                        action="store_true")

    args = parser.parse_args(args=args)

    if args.echo:
        print('\033[1m\033[31mExecuting: ' + ' '.join(sys.argv) + '\033[0m\n')

    if not args.overwrite and pathlib.Path(args.output).exists():
        print(f"Output file {args.output} already exists. Use --overwrite to overwrite it.")
        sys.exit(1)

    # compute operation
    compute_operation(
        file1=args.file1,
        file2=args.file2,
        operation=args.operation,
        extname1=args.extname1,
        extname2=args.extname2,
        output=args.output,
        overwrite=args.overwrite,
        dtype=args.dtype,
        display=args.display,
        args_z1z2=args.z1z2,
        args_bbox=args.bbox,
        args_keystitle=args.keystitle,
        args_geometry=args.geometry
    )


if __name__ == "__main__":

    main()



#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#
"""Check if all elements in a sequence are valid numbers (not NaN or Inf)."""

import math


def is_valid_number(x):
    """Check if x is a valid number (not NaN or Inf)."""
    return isinstance(x, (int, float)) and not math.isnan(x) and not math.isinf(x)


def all_valid_numbers(seq, fixed_length=None):
    """Check if all elements in seq are valid numbers.

    Parameters
    ----------
    seq : list or tuple
        Sequence of numbers to check.
    fixed_length : int or None, optional
        If provided, the sequence must have this exact length.

    Returns
    -------
    bool
        True if all elements are valid numbers (and length matches
        if fixed_length is set), False otherwise.
    """
    if not isinstance(seq, (list, tuple)):
        raise TypeError("Input must be a list or tuple.")
    if fixed_length is not None and len(seq) != fixed_length:
        print(f"Sequence {seq} has length {len(seq)}, expected {fixed_length}.")
        raise ValueError(f"Input sequence must have length {fixed_length}.")
    return all(is_valid_number(x) for x in seq)



#
# Copyright 2018-2026 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import os.path
import json
import uuid

import pytest

from numina.util.jsonencoder import ExtEncoder
from numina.tests.drptest import create_drp_test
from ..backend import Backend
from numina.exceptions import NoResultFound
import numina.types.qc as qc
import numina.instrument.assembly as asb
import numina.core.oresult
from numina.util.context import working_directory
import numina.tests.simpleobj as simpleobj


def repeat_my(result_content, result_name):
    import numina.store

    saveres = dict(qc=result_content["qc"], uuid=result_content["uuid"], values={})
    saveres_v = saveres["values"]

    result_values = result_content["values"]
    for key, obj in result_values.items():
        obj = result_values[key]
        saveres_v[key] = numina.store.dump(obj, obj, key)

    with open(result_name, "w") as fd:
        json.dump(saveres, fd, indent=2, cls=ExtEncoder)


@pytest.fixture
def backend(tmpdir):

    drps = create_drp_test(["drpclodia.yaml"])
    name = "CLODIA"
    ob_table = [
        dict(id=1, instrument=name, mode="sky", images=[], children=[], enabled=False),
        dict(id=2, instrument=name, mode="sky", images=[], children=[]),
        dict(id=3, instrument=name, mode="image", images=[], children=[]),
        dict(id=4, instrument=name, mode="sky", images=[], children=[]),
        dict(id=5, instrument=name, mode="image", images=[], children=[]),
        dict(id=30, instrument=name, mode="mosaic", images=[], children=[2, 3]),
        dict(id=40, instrument=name, mode="mosaic", images=[], children=[4, 5]),
        dict(id=400, instrument=name, mode="raiz", images=[], children=[30, 40]),
    ]

    prod_table = {
        "TEST1": {
            "225fcaf2-7f6f-49cc-972a-70fd0aee8e96": [
                {
                    "id": 1,
                    "type": "DemoType1",
                    "tags": {},
                    "content": {"demo1": 1},
                    "ob": 2,
                },
                {
                    "id": 2,
                    "type": "DemoType2",
                    "tags": {"field2": "A"},
                    "content": {"demo2": 2},
                    "ob": 14,
                },
                {
                    "id": 3,
                    "type": "DemoType2",
                    "tags": {"field2": "B"},
                    "content": {"demo2": 3},
                    "ob": 15,
                },
            ],
            "6ad5dc90-6b15-43b7-abb5-b07340e19f41": [
                {
                    "id": 33,
                    "type": "DemoType2",
                    "tags": {"field2": "B"},
                    "content": {"demo2": 9},
                    "ob": 25,
                }
            ],
        }
    }
    result_name = "result.json"
    results_table = {
        1: {
            "id": 1,
            "instrument": name,
            "mode": "image",
            "oblock_id": 5,
            "task_id": 1,
            "time_create": "2018-07-24T19:12:01",
            "result_dir": "dum1",
            "result_file": result_name,
            "qc": "GOOD",
        },
        2: {
            "id": 2,
            "instrument": name,
            "mode": "sky",
            "oblock_id": 4,
            "task_id": 2,
            "qc": "BAD",
            "time_create": "2018-07-24T19:12:09",
            "result_file": result_name,
            "result_dir": "dum2",
        },
        3: {
            "id": 3,
            "instrument": name,
            "mode": "image",
            "oblock_id": 5,
            "task_id": 3,
            "time_create": "2018-07-24T19:12:11",
            "qc": "GOOD",
            "result_dir": "dum3",
            "result_file": result_name,
        },
    }

    gentable = {}
    gentable["products"] = prod_table
    gentable["requirements"] = {}
    gentable["results"] = results_table
    # gentable['oblocks'] = ob_table
    # Load instrument profiles
    pkg_paths = ["numina.drps.tests.configs"]
    store = asb.load_paths_store(pkg_paths)

    result_name = "result.json"

    result1_dir = tmpdir.mkdir("dum1")
    result1_values = dict(
        calib=simpleobj.create_simple_structured(),
        reduced_rss=simpleobj.create_simple_frame(),
        reduced_image=simpleobj.create_simple_frame(),
    )
    result1_content = dict(
        qc="GOOD", values=result1_values, uuid="10000000-10000000-10000000-10000000"
    )
    with working_directory(str(result1_dir)):
        repeat_my(result1_content, result_name)

    result2_dir = tmpdir.mkdir("dum2")

    result2_values = dict(calib=simpleobj.create_simple_structured())
    result2_content = dict(
        qc="BAD", values=result2_values, uuid="20000000-20000000-20000000-20000000"
    )
    with working_directory(str(result2_dir)):
        repeat_my(result2_content, result_name)

    result3_dir = tmpdir.mkdir("dum3")

    result3_values = dict(
        reduced_rss="reduced_rss.fits", reduced_image="reduced_image.fits"
    )
    result3_content = dict(
        qc="BAD", values=result3_values, uuid="30000000-30000000-30000000-30000000"
    )
    with working_directory(str(result3_dir)):
        repeat_my(result3_content, result_name)

    base = Backend(drps, "", gentable, components=store, basedir=str(tmpdir))
    base.add_obs(ob_table)

    return base


@pytest.fixture
def backend_empty(tmpdir):
    drps = create_drp_test(["drpclodia.yaml"])
    pkg_paths = ["numina.drps.tests.configs"]
    store = asb.load_paths_store(pkg_paths)
    base = Backend(drps, "", {}, components=store, basedir=str(tmpdir))
    return base


def test_skip_reserved(backend):

    ss_ids = list(backend.search_session_ids())

    assert ss_ids == [2, 3, 4, 5, 30, 40]


def test_parent_inserted(backend):

    obsres = backend.search_oblock_from_id(2)
    assert obsres.parent == 30

    obsres = backend.search_oblock_from_id(4)
    assert obsres.parent == 40

    obsres = backend.search_oblock_from_id(30)
    assert obsres.parent == 400

    obsres = backend.search_oblock_from_id(400)
    assert obsres.parent is None


def test_previous_obsid(backend):

    obsres = backend.search_oblock_from_id(5)
    print(type(obsres))
    previd = backend.search_previous_obsres(obsres, node="prev")
    assert list(previd) == [4, 3, 2, 1]

    obsres = backend.search_oblock_from_id(5)
    previd = backend.search_previous_obsres(obsres, node="prev-rel")
    assert list(previd) == [4]

    obsres = backend.search_oblock_from_id(4)
    previd = backend.search_previous_obsres(obsres, node="prev-rel")
    assert list(previd) == []


def test_search_result_id(backend):
    from numina.types.frame import DataFrameType
    from numina.types.dataframe import DataFrame

    tipo = DataFrameType()
    node_id = 5
    field = "reduced_rss"
    res = backend.search_result_id(node_id, tipo, field, mode=None)

    assert isinstance(res.content, DataFrame)
    relpath = os.path.relpath(res.content.filename, backend.basedir)
    assert relpath == os.path.join("dum3", "reduced_rss.fits")


def test_search_result_id_notfound(backend):
    from numina.types.frame import DataFrameType

    tipo = DataFrameType()
    node_id = 2
    field = "reduced_rss"

    with pytest.raises(NoResultFound):
        backend.search_result_id(node_id, tipo, field, mode=None)


def test_build_recipe_result(backend):
    from numina.types.dataframe import DataFrame
    from numina.types.structured import BaseStructuredCalibration

    obj_uuid = "10000000-10000000-10000000-10000000"
    res = backend.build_recipe_result(result_id=1)

    assert res.qc == qc.QC.GOOD

    assert hasattr(res, "reduced_rss")
    assert isinstance(res.reduced_rss, DataFrame)

    assert hasattr(res, "reduced_image")
    assert isinstance(res.reduced_image, DataFrame)

    assert hasattr(res, "calib")
    assert isinstance(res.calib, BaseStructuredCalibration)

    assert res.uuid == uuid.UUID(obj_uuid)


def test_build_recipe_result2(backend):
    from numina.types.structured import BaseStructuredCalibration

    obj_uuid = "20000000-20000000-20000000-20000000"
    res = backend.build_recipe_result(result_id=2)

    assert res.qc == qc.QC.BAD

    assert hasattr(res, "calib")
    assert isinstance(res.calib, BaseStructuredCalibration)

    assert res.uuid == uuid.UUID(obj_uuid)


def test_ago(backend_empty):

    ob1 = {
        "id": 100,
        "instrument": "CLODIA",
        "mode": "image",
        "labels": {"obsid_wl": 400},
    }
    ob2 = {
        "id": 200,
        "instrument": "CLODIA",
        "mode": "image",
        "labels": {"obsid_wl": 400},
    }
    ob3 = {"id": 300, "instrument": "CLODIA", "mode": "image"}
    ob4 = {"id": 400, "instrument": "CLODIA", "mode": "image"}

    backend_empty.add_obs([ob1, ob2, ob3, ob4])

    oblock1 = backend_empty.oblock_from_id(100)
    assert isinstance(oblock1, numina.core.oresult.ObservingBlock)

    assert oblock1.labels == ob1["labels"]

    oblock2 = backend_empty.oblock_from_id(200)
    assert isinstance(oblock2, numina.core.oresult.ObservingBlock)

    obsres1 = backend_empty.obsres_from_oblock(oblock1)

    assert isinstance(obsres1, numina.core.oresult.ObservationResult)



#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Auxiliary script to perform binary image arithmetic with 3D cubes.

The computation is performed casting to np.float32."""

import argparse
from rich import print
from rich_argparse import RichHelpFormatter
import sys

from astropy.io import fits
import numpy as np
import pathlib

from .add_script_info_to_fits_history import add_script_info_to_fits_history


def compute_operation(file1, file2, operation, extname1, extname2, dtype):
    """Compute output = file1 operation file2.

    Parameters
    ----------
    file1 : file object
        First FITS file.
    file2 : file object or float
        Second FITS file or float number.
    operation : string
        Mathematical operation.
    extname1 : str
        Extension name of the first FITS file (default: 'PRIMARY').
    extname2 : str
        Extension name of the second FITS file (default: 'PRIMARY').
    dtype : str
        Data type of the output image.

    Returns
    -------
    solution : `numpy.ndarray`
        Resulting 3D image after applying the operation.
    """

    # read first FITS file
    with fits.open(file1) as hdulist:
        image_header1 = hdulist[extname1].header
        image1 = hdulist[extname1].data.astype(dtype)
    naxis = image_header1['naxis']
    if naxis != 3:
        raise ValueError("Input file must be a 3D cube (NAXIS=3).")
    naxis1 = image_header1['naxis1']
    naxis2 = image_header1['naxis2']
    naxis3 = image_header1['naxis3']

    # read second FITS file or number
    try:
        with fits.open(file2) as hdulist:
            image_header2 = hdulist[extname2].header
            image2 = hdulist[extname2].data.astype(dtype)
            naxis_ = image_header2['naxis']
            naxis1_ = image_header2['naxis1']
            naxis2_ = image_header2['naxis2']
            naxis3_ = image_header2['naxis3']
    except FileNotFoundError:
        image2 = np.zeros((naxis3, naxis2, naxis1), dtype=dtype)
        if 'int' in dtype:
            file2 = int(file2)
        elif 'float' in dtype:
            file2 = float(file2)
        else:
            raise ValueError(f"Unsupported dtype: {dtype}. Use an integer or float type.")
        image2 += file2  # if file2 is a number
        naxis_ = naxis
        naxis1_ = naxis1
        naxis2_ = naxis2
        naxis3_ = naxis3

    # additional dimension checks
    if naxis != naxis_:
        raise ValueError("NAXIS values are different.")
    if naxis1 != naxis1_:
        raise ValueError("NAXIS1 values are different.")
    if naxis2 != naxis2_:
        raise ValueError("NAXIS2 values are different.")
    if naxis3 != naxis3_:
        raise ValueError("NAXIS3 values are different.")

    # compute operation
    if operation == "+":
        solution = image1 + image2
    elif operation == "-":
        solution = image1 - image2
    elif operation == "x":
        solution = image1 * image2
    elif operation == "/":
        solution = image1 / image2
    elif operation == "=":
        solution = image2.copy()
    else:
        raise ValueError("Unexpected operation=" + str(operation))

    return solution


def main(args=None):

    # parse command-line options
    parser = argparse.ArgumentParser(
        description="description: binary image arithmetic",
        formatter_class=RichHelpFormatter
    )
    # positional parameters
    parser.add_argument("file1",
                        help="First FITS image",
                        type=argparse.FileType('rb'))
    parser.add_argument("operation",
                        help="Arithmetic operation",
                        type=str,
                        choices=['+', '-', 'x', '/', '='])
    parser.add_argument("file2",
                        help="Second FITS image or number",
                        type=str)
    parser.add_argument("output",
                        help="Output FITS image",
                        type=str)
    # optional arguments
    parser.add_argument("--extname1", type=str,
                        help="Extension name of the first FITS file (default: 'PRIMARY').",
                        default='PRIMARY')
    parser.add_argument("--extname2", type=str,
                        help="Extension name of the second FITS file (default: 'PRIMARY').",
                        default='PRIMARY')
    parser.add_argument("--overwrite",
                        help="Overwrite output file if already exists",
                        action="store_true")
    parser.add_argument("--dtype",
                        help="Data type of the output image (default: float32)",
                        type=str,
                        choices=['uint8', 'int8', 'uint16', 'int16', 'uint32', 'int32', 'uint64', 'int64',
                                 'float32', 'float64'],
                        default='float32')
    parser.add_argument("--echo",
                        help="Display full command line",
                        action="store_true")

    args = parser.parse_args(args=args)

    if args.echo:
        print('\033[1m\033[31mExecuting: ' + ' '.join(sys.argv) + '\033[0m\n')

    if not args.overwrite and pathlib.Path(args.output).exists():
        print(f"Output file {args.output} already exists. Use --overwrite to overwrite it.")
        sys.exit(1)

    # compute operation
    solution = compute_operation(
        file1=args.file1,
        file2=args.file2,
        operation=args.operation,
        extname1=args.extname1,
        extname2=args.extname2,
        dtype=args.dtype
    )

    # save output file
    header = fits.getheader(args.file1, args.extname1)
    if header['naxis1'] != solution.shape[2] or \
       header['naxis2'] != solution.shape[1] or \
       header['naxis3'] != solution.shape[0]:
        raise ValueError("Output image dimensions do not match the input image dimensions.")
    hdu = fits.PrimaryHDU(solution.astype(args.dtype), header=header)
    add_script_info_to_fits_history(hdu.header, args)
    hdu.writeto(args.output, overwrite=args.overwrite)


if __name__ == "__main__":

    main()



#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Progress bar that logs complete lines."""


class ProgressBarLines:
    """Progress bar that logs complete lines."""

    def __init__(self, total=None, logger=None):
        if total is None or logger is None:
            raise ValueError("Total and logger must be provided")
        self.total = total
        self.current = 0
        self.shown_milestones = set([0])
        self.logger = logger
        self.progress_line = "0%"
        self.logger.info(self.progress_line)

    def update(self, step=1):
        """Update progress."""
        self.current += step
        percent = (self.current / self.total) * 100

        for milestone in range(10, 101, 10):
            if percent >= milestone and milestone not in self.shown_milestones:
                self.shown_milestones.add(milestone)
                self.progress_line += f" {milestone}%"
                self.logger.info(self.progress_line)



#
# Copyright 2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#
"""Compute the average excluding the top N values."""

import numpy as np


def average_excluding_top_n(arr, n, axis=0):
    """
    Compute average along specified axis, excluding the N largest values.

    Note that np.partition is more efficient than full sorting
    when you only need to separate the N largest values.
    The function handles arbitrary axis selection.
    Time complexity is O(n) average case, versus O(n log n) for
    full sorting

    Parameters
    ----------
    arr : ndarray
        Input array
    n : int
        Number of largest values to exclude
    axis : int, optional
        Axis along which to compute the average (default: 0)

    Returns
    -------
    ndarray
        Array with reduced dimensionality along specified axis
    """
    if n >= arr.shape[axis]:
        raise ValueError(f"n ({n}) must be smaller than array size along axis {axis} ({arr.shape[axis]})")

    # Use partition to find the (n+1)th largest value
    # Everything before this index will be the smaller values
    kth = arr.shape[axis] - n

    # Partition the array along the specified axis
    partitioned = np.partition(arr, kth, axis=axis)

    # Take only the values up to (not including) the N largest
    slices = [slice(None)] * arr.ndim
    slices[axis] = slice(None, kth)

    return np.mean(partitioned[tuple(slices)], axis=axis)



#
# Copyright 2024-2026 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

from astropy.io import fits
import matplotlib.pyplot as plt

from .raise_valueerror import raise_ValueError


def display_skycalc(faux_skycalc):
    """
    Display sky radiance and transmission.

    Data generated with the SKYCALC Sky Model Calculator tool
    provided by ESO.
    See https://www.eso.org/observing/etc/doc/skycalc/helpskycalc.html

    Parameters
    ----------
    faux_skycalc : str
        FITS file name with SKYCALC predictions.
    """

    with fits.open(faux_skycalc) as hdul:
        skycalc_header = hdul[1].header
        skycalc_table = hdul[1].data

    if skycalc_header["TTYPE1"] != "lam":
        raise_ValueError(f"Unexpected TTYPE1: {skycalc_header['TTYPE1']}")
    if skycalc_header["TTYPE2"] != "flux":
        raise_ValueError(f"Unexpected TTYPE2: {skycalc_header['TTYPE2']}")

    wave = skycalc_table["lam"]
    flux = skycalc_table["flux"]
    trans = skycalc_table["trans"]

    # plot radiance
    fig, ax = plt.subplots()
    ax.plot(wave, flux, "-", linewidth=1)
    cwave_unit = skycalc_header["TUNIT1"]
    cflux_unit = skycalc_header["TUNIT2"]
    ax.set_xlabel(f"Vacuum Wavelength ({cwave_unit})")
    ax.set_ylabel(f"{cflux_unit}")
    ax.set_title("SKYCALC prediction")
    plt.tight_layout()
    plt.show()

    # plot transmission
    fig, ax = plt.subplots()
    ax.plot(wave, trans, "-", linewidth=1)
    ax.set_xlabel(f"Vacuum Wavelength ({cwave_unit})")
    ax.set_ylabel("Transmission fraction")
    ax.set_title("SKYCALC prediction")
    plt.tight_layout()
    plt.show()



#
# Copyright 2025-2026 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Display 2D histogram of min2d-bias vs mean2d-bias values."""

from astropy.table import Table
from matplotlib.colors import LogNorm
from matplotlib.colors import LinearSegmentedColormap
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from pathlib import Path
from scipy import ndimage
import sys

from numina.array.numsplines import spline_positive_derivative
from numina.tools.input_number import input_number
import teareduce as tea

from .apply_threshold_cr import apply_threshold_cr
from .define_piecewise_linear_function import define_piecewise_linear_function
from .valid_parameters import VALID_BOUNDARY_FITS
from .valid_parameters import DEFAULT_WEIGHT_FIXED_POINTS_IN_BOUNDARY

mpl.rcParams["keymap.quit"] = []  # disable 'q' key for quitting plots
mpl.rcParams["keymap.fullscreen"] = []  # disable 'f' key for fullscreen (used here for other purpose)


def xsort_and_show_fixed_points_in_boundary(
    _logger,
    num_fixed_points,
    x_mm_fixed_points_in_boundary,
    y_mm_fixed_points_in_boundary,
    w_mm_fixed_points_in_boundary,
):
    """Show fixed points in boundary.

    If there are fixed points, sort them by increasing x value.
    Return list of fixed points or None if there are no fixed points.
    """
    if num_fixed_points > 0:
        isort = np.argsort(x_mm_fixed_points_in_boundary)
        if not np.all(isort == np.arange(num_fixed_points)):
            _logger.info("Sorting fixed points in boundary by increasing x value.")
        x_mm_fixed_points_in_boundary[:] = x_mm_fixed_points_in_boundary[isort]
        y_mm_fixed_points_in_boundary[:] = y_mm_fixed_points_in_boundary[isort]
        w_mm_fixed_points_in_boundary[:] = w_mm_fixed_points_in_boundary[isort]
        fixed_table = Table(names=("number", "X", "Y", "Weight"), dtype=(int, float, float, float))
        mm_fixed_points_in_boundary = []
        for idum in range(num_fixed_points):
            fixed_table.add_row(
                (
                    idum + 1,
                    x_mm_fixed_points_in_boundary[idum],
                    y_mm_fixed_points_in_boundary[idum],
                    w_mm_fixed_points_in_boundary[idum],
                )
            )
            mm_fixed_points_in_boundary.append(
                [
                    x_mm_fixed_points_in_boundary[idum],
                    y_mm_fixed_points_in_boundary[idum],
                    w_mm_fixed_points_in_boundary[idum],
                ]
            )
        _logger.info("Current fixed points in boundary:\n%s", fixed_table)
    else:
        _logger.info("No fixed points in boundary.")
        mm_fixed_points_in_boundary = None

    return mm_fixed_points_in_boundary


def display_hist2d(
    _logger,
    rlabel_mmcosmic,
    mm_hist2d_min_neighbors,
    hist2d_accummulated,
    mm_nsimulations,
    mm_synthetic,
    mm_photon_distribution,
    mm_nbinom_shape,
    bins_xdiag,
    bins_ydiag,
    xplot,
    yplot,
    xdiag_min,
    xdiag_max,
    max2d,
    bool_to_be_cleaned,
    rnoise,
    mm_threshold,
    mm_boundary_fit,
    mm_knots_splfit,
    mm_minimum_max2d_rnoise,
    mm_dilation,
    mm_niter_boundary_extension,
    mm_weight_boundary_extension,
    mm_fixed_points_in_boundary,
    x_mm_fixed_points_in_boundary,
    y_mm_fixed_points_in_boundary,
    w_mm_fixed_points_in_boundary,
    interactive,
    output_dir=".",
    record_terminal_output=False,
):
    """Display 2D histogram of min2d-bias vs mean2d-bias values."""
    # Remove bins that are surrounded by less than mm_hist2d_min_neighbors neighbors
    hist2d_accummulated_original = hist2d_accummulated.copy()
    loop = True
    while loop:
        hist2d_accummulated = hist2d_accummulated_original.copy()
        if mm_hist2d_min_neighbors > 0:
            kernel_neighbors = np.array([[1, 1, 1], [1, 0, 1], [1, 1, 1]], dtype=int)
            neighbor_counts = ndimage.convolve(
                (hist2d_accummulated > 0).astype(int), kernel_neighbors, mode="constant", cval=0
            )
            surrounded_by_threshold = (hist2d_accummulated > 0) & (neighbor_counts < mm_hist2d_min_neighbors)
            num_zero_bins_before = np.sum(hist2d_accummulated == 0)
            hist2d_accummulated[surrounded_by_threshold] = 0
            num_zero_bins_after = np.sum(hist2d_accummulated == 0)
            _logger.info(
                "removed %d bins that were surrounded by less than %d neighbors",
                num_zero_bins_after - num_zero_bins_before,
                mm_hist2d_min_neighbors,
            )
        # Average the histogram over the number of simulations
        hist2d_accummulated = hist2d_accummulated.astype(float) / mm_nsimulations
        # Determine vmin and vmax for the color scale
        vmin = np.min(hist2d_accummulated[hist2d_accummulated > 0])
        if vmin == 0:
            vmin = 1
        vmax = np.max(hist2d_accummulated)
        cmap1 = plt.get_cmap("cividis_r")
        cmap2 = plt.get_cmap("viridis")
        n_colors = 256
        n_colors2 = int((np.log10(vmax) - np.log10(1.0)) / (np.log10(vmax) - np.log10(vmin)) * n_colors)
        n_colors2 += 1
        if n_colors2 > n_colors:
            n_colors2 = n_colors
        if n_colors2 < n_colors:
            n_colors1 = n_colors - n_colors2
        else:
            n_colors1 = 0
        colors1 = cmap1(np.linspace(0, 1, n_colors1))
        colors2 = cmap2(np.linspace(0, 1, n_colors2))
        combined_colors = np.vstack((colors1, colors2))
        combined_cmap = LinearSegmentedColormap.from_list("combined_cmap", combined_colors)
        norm = LogNorm(vmin=vmin, vmax=vmax)

        def on_key_2dhist(event):
            nonlocal loop, rerun_simulations
            if event.key == "x":
                _logger.info("Exiting program as per user request ('x' key pressed).")
                plt.close(fig)
                sys.exit(0)
            elif event.key == "c":
                _logger.info("Continuing to next step as per user request ('c' key pressed).")
                plt.close(fig)
                loop = False
                rerun_simulations = False
            elif event.key == "f":
                _logger.info("Repeating fit as per user request ('f' key pressed).")
                plt.close(fig)
                rerun_simulations = False
            elif event.key == "r":
                _logger.info("Rerunning simulations and fit as per user request ('r' key pressed).")
                plt.close(fig)
                loop = False
                rerun_simulations = True

        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12.1, 5.5))
        fig.canvas.mpl_connect("key_press_event", lambda event: on_key_2dhist(event))
        # Display 2D histogram of the simulated data
        extent = [bins_xdiag[0], bins_xdiag[-1], bins_ydiag[0], bins_ydiag[-1]]
        tea.imshow(
            fig,
            ax1,
            hist2d_accummulated,
            norm=norm,
            extent=extent,
            aspect="auto",
            cblabel="Number of pixels",
            cmap=combined_cmap,
        )
        # Display 2D histogram of the original data
        hist2d_original, edges = np.histogramdd(sample=(yplot, xplot), bins=(bins_ydiag, bins_xdiag))
        tea.imshow(
            fig,
            ax2,
            hist2d_original,
            norm=norm,
            extent=extent,
            aspect="auto",
            cblabel="Number of pixels",
            cmap=combined_cmap,
        )

        # Determine the detection boundary for coincident cosmic-ray detection
        _logger.info("computing numerical boundary for coincident cosmic-ray detection...")
        xboundary = []
        yboundary = []
        xcbins = (bins_xdiag[:-1] + bins_xdiag[1:]) / 2
        ycbins = (bins_ydiag[:-1] + bins_ydiag[1:]) / 2
        nbins_xdiag = len(xcbins)
        nbins_ydiag = len(ycbins)
        minimum_bin_value = 1.0 / mm_nsimulations
        for i in range(nbins_xdiag):
            fsum = np.sum(hist2d_accummulated[:, i])
            if fsum > 0:
                xboundary.append(xcbins[i])
                # Compute the probability density function for this x bin
                pdensity = hist2d_accummulated[:, i] / fsum
                # Find the y value where the cumulative distribution reaches the desired percentile
                # (note that 1 / mm_nsimulations is the minimum expected value different from zero in any bin)
                if fsum > minimum_bin_value:
                    perc = (fsum - minimum_bin_value) / fsum
                    # Interpolate to find the corresponding y value
                    p = np.interp(perc, np.cumsum(pdensity), np.arange(nbins_ydiag))
                    yboundary.append(ycbins[int(p + 0.5)])
                else:
                    yboundary.append(bins_ydiag[0])
        xboundary = np.array(xboundary)
        yboundary = np.array(yboundary)
        ax1.plot(xboundary, yboundary, "r+", label="Boundary points")
        boundaryfit = None  # avoid flake8 warning
        if mm_boundary_fit == "spline":
            nmax_iterations_with_color = 6
            for iterboundary in range(mm_niter_boundary_extension + 1):
                wboundary = np.ones_like(xboundary, dtype=float)
                color = f"C{iterboundary}"
                alpha = 1.0
                if iterboundary == 0:
                    label = "Initial spline fit"
                else:
                    wboundary[yboundary > boundaryfit(xboundary)] = mm_weight_boundary_extension**iterboundary
                    if iterboundary == mm_niter_boundary_extension:
                        label = f"Final iteration {iterboundary}"
                        if mm_niter_boundary_extension > nmax_iterations_with_color:
                            color = f"C{nmax_iterations_with_color}"
                    else:
                        if mm_niter_boundary_extension > nmax_iterations_with_color:
                            if iterboundary < nmax_iterations_with_color:
                                label = f"Iteration {iterboundary}"
                            elif iterboundary == nmax_iterations_with_color:
                                if nmax_iterations_with_color == mm_niter_boundary_extension - 1:
                                    label = f"Iteration {iterboundary}"
                                else:
                                    label = (
                                        f"Iterations {nmax_iterations_with_color} to {mm_niter_boundary_extension - 1}"
                                    )
                                color = "gray"
                                alpha = 0.3
                            else:
                                label = None
                                color = "gray"
                                alpha = 0.3
                        else:
                            label = f"Iteration {iterboundary}"
                if mm_fixed_points_in_boundary is None:
                    xboundary_fit = xboundary
                    yboundary_fit = yboundary
                    wboundary_fit = wboundary
                else:
                    wboundary_max = np.max(wboundary)
                    xboundary_fit = np.concatenate((xboundary, x_mm_fixed_points_in_boundary))
                    yboundary_fit = np.concatenate((yboundary, y_mm_fixed_points_in_boundary))
                    wboundary_fit = np.concatenate((wboundary, w_mm_fixed_points_in_boundary * wboundary_max))
                isort = np.argsort(xboundary_fit)
                boundaryfit, knots = spline_positive_derivative(
                    x=xboundary_fit[isort],
                    y=yboundary_fit[isort],
                    w=wboundary_fit[isort],
                    n_total_knots=mm_knots_splfit,
                )
                ydum = boundaryfit(xcbins)
                ydum[xcbins < knots[0]] = boundaryfit(knots[0])
                ydum[xcbins > knots[-1]] = boundaryfit(knots[-1])
                ax1.plot(xcbins, ydum, "-", color=color, label=label)
                ax1.plot(knots, boundaryfit(knots), "o", color=color, alpha=alpha, markersize=4)
        elif mm_boundary_fit == "piecewise":
            boundaryfit = define_piecewise_linear_function(
                xarray=x_mm_fixed_points_in_boundary, yarray=y_mm_fixed_points_in_boundary
            )
            ax1.plot(xcbins, boundaryfit(xcbins), "r-", label="Piecewise linear fit")
        else:
            raise ValueError(f"Invalid mm_boundary_fit: {mm_boundary_fit}. Valid options are {VALID_BOUNDARY_FITS}.")

        if mm_threshold is None:
            # Use the minimum value of the boundary as the mm_threshold
            mm_threshold = np.min(yplot_boundary)
            _logger.info("updated mm_threshold for cosmic-ray detection: %f", mm_threshold)

        # Apply the criterium to detect coincident cosmic-ray pixels
        flag1 = yplot > boundaryfit(xplot)
        flag2 = yplot > mm_threshold
        flag_mm = np.logical_and(flag1, flag2)
        flag3 = max2d.flatten() > mm_minimum_max2d_rnoise * rnoise.flatten()
        flag_mm = np.logical_and(flag_mm, flag3)
        if mm_dilation > 0:
            _logger.info("applying binary dilation with size=%d to cosmic-ray mask", mm_dilation)
            num_pixels_before_dilation = np.sum(flag_mm)
            structure = ndimage.generate_binary_structure(2, 2)
            naxis2, naxis1 = max2d.shape
            flag_mm = ndimage.binary_dilation(
                flag_mm.reshape((naxis2, naxis1)), structure=structure, iterations=mm_dilation
            ).flatten()
            num_pixels_after_dilation = np.sum(flag_mm)
            ldum = len(str(num_pixels_after_dilation))
            _logger.info(f"number of pixels flagged before dilation : {num_pixels_before_dilation:{ldum}d}")
            _logger.info(f"number of pixels flagged after dilation  : {num_pixels_after_dilation:{ldum}d}")
            # Apply the thresholding again after dilation
            _logger.info("applying thresholding again after dilation")
            flag_mm = apply_threshold_cr(
                bool_crmask2d=flag_mm.reshape((naxis2, naxis1)), bool_threshold2d=flag2.reshape((naxis2, naxis1))
            ).flatten()
            num_pixels_after_thresholding = np.sum(flag_mm)
            _logger.info(f"number of pixels after thresholding again: {num_pixels_after_thresholding:{ldum}d}")
        # Apply the cleaning mask
        _logger.info("applying cleaning mask region")
        flag_mm = np.logical_and(flag_mm, bool_to_be_cleaned.flatten())
        _logger.info(
            "pixels flagged as cosmic rays by %s: %d (%08.4f%%)",
            rlabel_mmcosmic,
            np.sum(flag_mm),
            np.sum(flag_mm) / flag_mm.size * 100,
        )

        # Plot the results
        if mm_fixed_points_in_boundary is not None:
            ax1.plot(
                x_mm_fixed_points_in_boundary,
                y_mm_fixed_points_in_boundary,
                "ms",
                markersize=6,
                alpha=0.5,
                label="Fixed points",
            )
        ax1.set_xlabel(r"min2d $-$ bias  [ADU]")
        ax1.set_ylabel(r"median2d $-$ min2d  [ADU]")
        tmp_title = "Simulated data, "
        if mm_photon_distribution == "poisson":
            tmp_title += "Poisson"
        elif mm_photon_distribution == "nbinom":
            tmp_title += f"Negative binomial (shape={mm_nbinom_shape})"
        tmp_title += f"\n(mm_nsimulations = {mm_nsimulations}, mm_synthetic={mm_synthetic})"
        ax1.set_title(tmp_title)
        if mm_niter_boundary_extension > 1:
            ax1.legend(loc=1)
        xplot_boundary = np.linspace(xdiag_min, xdiag_max, 100)
        yplot_boundary = boundaryfit(xplot_boundary)
        if mm_boundary_fit == "spline":
            # For spline fit, force the boundary to be constant outside the knots
            yplot_boundary[xplot_boundary < knots[0]] = boundaryfit(knots[0])
            yplot_boundary[xplot_boundary > knots[-1]] = boundaryfit(knots[-1])
        ax2.plot(xplot_boundary, yplot_boundary, "r-", label="Detection boundary")
        if mm_fixed_points_in_boundary is not None:
            ax2.plot(
                x_mm_fixed_points_in_boundary,
                y_mm_fixed_points_in_boundary,
                "ms",
                markersize=6,
                alpha=0.5,
                label="Fixed points",
            )
        ax2.set_xlim(xdiag_min, xdiag_max)
        ax2.set_ylim(ax1.get_ylim())
        ax2.set_xlabel(ax1.get_xlabel())
        ax2.set_ylabel(ax1.get_ylabel())
        ax2.set_title("Original data")
        ax2.legend(loc=1)
        plt.tight_layout()
        png_filename = "diagnostic_histogram2d.png"
        _logger.info(f"saving {png_filename}")
        plt.savefig(Path(output_dir) / png_filename, dpi=150)
        if interactive:
            _logger.info(
                "Entering interactive mode (press any of the following keys)\n"
                "'f' to repeat fit, 'r' to rerun simulations, 'c' to continue, 'x' to quit"
            )
            plt.show()
        else:
            loop = False
            rerun_simulations = False
        if loop:
            # Modify hist2d_min_neighbors
            mm_hist2d_min_neighbors_ = mm_hist2d_min_neighbors
            prompt = "Minimum number of neighbors to keep bins in the 2D histogram (0-8)"
            mm_hist2d_min_neighbors = input_number(
                expected_type="int",
                prompt=prompt,
                min_val=0,
                max_val=8,
                default=mm_hist2d_min_neighbors,
            )
            if record_terminal_output:
                _logger.info(f"{prompt} [{mm_hist2d_min_neighbors_}]: {mm_hist2d_min_neighbors}")
            # Modify mm_boundary_fit
            new_mm_boundary_fit = ""
            mm_boundary_fit_ = mm_boundary_fit
            while new_mm_boundary_fit not in ["piecewise", "spline"]:
                prompt = "Type of boundary fit (piecewise | spline)"
                new_mm_boundary_fit = input(f"{prompt} [{mm_boundary_fit}]: ").strip().lower()
                if new_mm_boundary_fit == "":
                    new_mm_boundary_fit = mm_boundary_fit
                if new_mm_boundary_fit not in ["piecewise", "spline"]:
                    _logger.info("Invalid value. Please type 'piecewise' or 'spline'.")
            mm_boundary_fit = new_mm_boundary_fit
            if record_terminal_output:
                # escape first bracket for rich logging
                _logger.info(f"{prompt} \\[{mm_boundary_fit_}]: {mm_boundary_fit}")
            # Modify spline parameters if needed
            if mm_boundary_fit == "spline":
                # Modify number of knots
                mm_knots_splfit_ = mm_knots_splfit
                prompt = "Number of knots for spline boundary fit (min. 2)"
                mm_knots_splfit = input_number(
                    expected_type="int",
                    prompt=prompt,
                    min_val=2,
                    default=mm_knots_splfit,
                )
                if record_terminal_output:
                    _logger.info(f"{prompt} [{mm_knots_splfit_}]: {mm_knots_splfit}")
                # Modify number of iterations for boundary extension
                mm_niter_boundary_extension_ = mm_niter_boundary_extension
                prompt = "Number of iterations for boundary extension (min. 0)"
                mm_niter_boundary_extension = input_number(
                    expected_type="int",
                    prompt=prompt,
                    min_val=0,
                    default=mm_niter_boundary_extension,
                )
                if record_terminal_output:
                    _logger.info(f"{prompt} [{mm_niter_boundary_extension_}]: {mm_niter_boundary_extension}")
                # Modify weight for boundary extension
                mm_weight_boundary_extension_ = mm_weight_boundary_extension
                prompt = "Weight for boundary extension (greater than 1.0)"
                mm_weight_boundary_extension = input_number(
                    expected_type="float",
                    prompt=prompt,
                    min_val=1.0,
                    default=mm_weight_boundary_extension,
                )
                if record_terminal_output:
                    _logger.info(f"{prompt} [{mm_weight_boundary_extension_}]: {mm_weight_boundary_extension}")
            # Modify fixed points in the boundary
            if mm_fixed_points_in_boundary is None:
                num_fixed_points = 0
            else:
                num_fixed_points = len(x_mm_fixed_points_in_boundary)
            if mm_boundary_fit == "piecewise" and num_fixed_points < 2:
                _logger.info("At least two fixed points are needed for piecewise linear boundary fit.")
                modify_fixed = "y"
            else:
                modify_fixed = "?"
            while modify_fixed != "n":
                mm_fixed_points_in_boundary = xsort_and_show_fixed_points_in_boundary(
                    _logger,
                    num_fixed_points,
                    x_mm_fixed_points_in_boundary,
                    y_mm_fixed_points_in_boundary,
                    w_mm_fixed_points_in_boundary,
                )
                while modify_fixed not in ["y", "n", ""]:
                    prompt = "Do you want to modify the fixed points in the boundary"
                    modify_fixed = input(f"{prompt} (y/[n])? ").strip().lower()
                    if modify_fixed not in ["y", "n", ""]:
                        _logger.info("Invalid value. Please type 'y' or 'n' or press 'Enter'.")
                if record_terminal_output:
                    _logger.info(f"{prompt} (y/\\[n])? {modify_fixed}")
                if modify_fixed != "y":
                    modify_fixed = "n"
                # Allow to delete individual fixed points or add new ones
                if modify_fixed == "y":
                    action = ""
                    while action != "n":
                        submenu = "Type:\n"
                        submenu += "- 'a' to add a fixed point\n"
                        valid_answers = "a/[n]"
                        valid_answers_ = "a/\\[n]"
                        if num_fixed_points > 0:
                            submenu += "- 'c' to clear all fixed points\n"
                            submenu += "- 'd' to delete an existing fixed point\n"
                            submenu += "- 'e' to edit an existing fixed point\n"
                            valid_answers = "a/c/d/e/[n]"
                            valid_answers_ = "a/c/d/e/\\[n]"
                        submenu += "- 'n' none (continue without additional changes)"
                        _logger.info(submenu)
                        prompt = f"Your choice"
                        action = input(f"{prompt} ({valid_answers}): ").strip().lower()
                        if record_terminal_output:
                            _logger.info(f"{prompt} ({valid_answers_}): {action}")
                        if action == "d" and num_fixed_points > 0:
                            if num_fixed_points == 1:
                                _logger.info("Only one fixed point available, deleting it.")
                                index_to_delete = 1
                            else:
                                prompt = f"Index of fixed point to delete (1 to {num_fixed_points})"
                                index_to_delete = input_number(
                                    expected_type="int",
                                    prompt=prompt,
                                    min_val=1,
                                    max_val=num_fixed_points,
                                )
                                if record_terminal_output:
                                    _logger.info(f"{prompt}: {index_to_delete}")
                            x_mm_fixed_points_in_boundary = np.delete(
                                x_mm_fixed_points_in_boundary, index_to_delete - 1
                            )
                            y_mm_fixed_points_in_boundary = np.delete(
                                y_mm_fixed_points_in_boundary, index_to_delete - 1
                            )
                            w_mm_fixed_points_in_boundary = np.delete(
                                w_mm_fixed_points_in_boundary, index_to_delete - 1
                            )
                            num_fixed_points -= 1
                            if num_fixed_points == 0:
                                x_mm_fixed_points_in_boundary = None
                                y_mm_fixed_points_in_boundary = None
                                w_mm_fixed_points_in_boundary = None
                        elif action == "e" and num_fixed_points > 0:
                            if num_fixed_points == 1:
                                _logger.info("Only one fixed point available, editing it.")
                                index_to_edit = 1
                            else:
                                prompt = f"Index of fixed point to edit (1 to {num_fixed_points})"
                                index_to_edit = input_number(
                                    expected_type="int",
                                    prompt=prompt,
                                    min_val=1,
                                    max_val=num_fixed_points,
                                )
                                if record_terminal_output:
                                    _logger.info(f"{prompt}: {index_to_edit}")
                            prompt = "New x value of fixed point"
                            x_new = input_number(
                                expected_type="float",
                                prompt=prompt,
                                default=x_mm_fixed_points_in_boundary[index_to_edit - 1],
                            )
                            if record_terminal_output:
                                _logger.info(f"{prompt} [{x_mm_fixed_points_in_boundary[index_to_edit - 1]}]: {x_new}")
                            prompt = "New y value of fixed point"
                            y_new = input_number(
                                expected_type="float",
                                prompt=prompt,
                                default=y_mm_fixed_points_in_boundary[index_to_edit - 1],
                            )
                            if record_terminal_output:
                                _logger.info(f"{prompt} [{y_mm_fixed_points_in_boundary[index_to_edit - 1]}]: {y_new}")
                            prompt = "New weight of fixed point"
                            w_new = input_number(
                                expected_type="float",
                                prompt=prompt,
                                min_val=0.0,
                                default=w_mm_fixed_points_in_boundary[index_to_edit - 1],
                            )
                            if record_terminal_output:
                                _logger.info(f"{prompt} [{w_mm_fixed_points_in_boundary[index_to_edit - 1]}]: {w_new}")
                            x_mm_fixed_points_in_boundary[index_to_edit - 1] = x_new
                            y_mm_fixed_points_in_boundary[index_to_edit - 1] = y_new
                            w_mm_fixed_points_in_boundary[index_to_edit - 1] = w_new
                        elif action == "a":
                            prompt = "x value of new fixed point"
                            x_new = input_number(expected_type="float", prompt=prompt, default=None)
                            if record_terminal_output:
                                _logger.info(f"{prompt}: {x_new}")
                            prompt = "y value of new fixed point"
                            y_new = input_number(expected_type="float", prompt=prompt, default=None)
                            if record_terminal_output:
                                _logger.info(f"{prompt}: {y_new}")
                            prompt = "weight of new fixed point"
                            w_new = input_number(
                                expected_type="float",
                                prompt=prompt,
                                min_val=0.0,
                                default=DEFAULT_WEIGHT_FIXED_POINTS_IN_BOUNDARY,
                            )
                            if record_terminal_output:
                                _logger.info(f"{prompt} [{DEFAULT_WEIGHT_FIXED_POINTS_IN_BOUNDARY}]: {w_new}")
                            if num_fixed_points == 0:
                                x_mm_fixed_points_in_boundary = np.array([x_new], dtype=float)
                                y_mm_fixed_points_in_boundary = np.array([y_new], dtype=float)
                                w_mm_fixed_points_in_boundary = np.array([w_new], dtype=float)
                            else:
                                x_mm_fixed_points_in_boundary = np.append(x_mm_fixed_points_in_boundary, x_new)
                                y_mm_fixed_points_in_boundary = np.append(y_mm_fixed_points_in_boundary, y_new)
                                w_mm_fixed_points_in_boundary = np.append(w_mm_fixed_points_in_boundary, w_new)
                            num_fixed_points += 1
                        elif action == "c" and num_fixed_points > 0:
                            num_fixed_points = 0
                            x_mm_fixed_points_in_boundary = None
                            y_mm_fixed_points_in_boundary = None
                            w_mm_fixed_points_in_boundary = None
                        elif action in ["n", ""]:
                            if action == "":
                                action = "n"
                            if mm_boundary_fit == "piecewise":
                                if num_fixed_points < 2:
                                    _logger.info(
                                        "At least two fixed points are needed for piecewise linear boundary fit."
                                    )
                                    prompt = "Press Enter to continue..."
                                    input(prompt)
                                    if record_terminal_output:
                                        _logger.info(prompt)
                                    action = "?"
                            if action == "n":
                                _logger.info("No changes made to fixed points in boundary.")
                                modify_fixed = "n"
                        else:
                            input("Invalid option. Press Enter to try again...")
                        mm_fixed_points_in_boundary = xsort_and_show_fixed_points_in_boundary(
                            _logger,
                            num_fixed_points,
                            x_mm_fixed_points_in_boundary,
                            y_mm_fixed_points_in_boundary,
                            w_mm_fixed_points_in_boundary,
                        )
        else:
            plt.close(fig)

    result = {
        "boundaryfit": boundaryfit,
        "flag_mm": flag_mm,
        "mm_fixed_points_in_boundary": mm_fixed_points_in_boundary,
        "mm_hist2d_min_neighbors": mm_hist2d_min_neighbors,
        "mm_boundary_fit": mm_boundary_fit,
        "mm_knots_splfit": mm_knots_splfit,
        "mm_niter_boundary_extension": mm_niter_boundary_extension,
        "mm_weight_boundary_extension": mm_weight_boundary_extension,
    }

    return rerun_simulations, result



#
# Copyright 2024-2026 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#
import logging

from astropy.io import fits
from astropy.units import Quantity
import astropy.units as u
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
import numpy as np
from pathlib import Path

from .define_3d_wcs import wcs_to_header_using_cd_keywords


def generate_image2d_method0_ifu(
    wcs3d,
    header_keys,
    noversampling_whitelight,
    simulated_x_ifu_all,
    simulated_y_ifu_all,
    prefix_intermediate_fits,
    instname,
    subtitle,
    scene,
    plots,
    logger=None,
    output_dir=".",
):
    """Compute image2d IFU (white image), method0

    The image is calculated by performing a two-dimensional histogram
    over the X, Y coordinates of the simulated photons in the IFU,
    regardless of the wavelength assigned to each photon.

    The result is saved as a FITS file.

    Parameters
    ----------
    wcs3d : `~astropy.wcs.wcs.WCS`
        WCS of the data cube.
    header_keys : `~astropy.io.fits.header.Header`
        FITS header with additional keywords to be merged together with
        the WCS information.
    noversampling_whitelight : int
        Oversampling factor (integer number) to generate the method0
        white image.
    simulated_x_ifu_all : `~astropy.units.Quantity`
        Simulated X coordinates of the photons in the IFU.
    simulated_y_ifu_all : `~astropy.units.Quantity`
        Simulated Y coordinates of the photons in the IFU.
    prefix_intermediate_fits : str
        Prefix for output intermediate FITS files. If the length of
        this string is 0, no output is generated.
    instname : str or None
        Instrument name.
    subtitle : str or None
        Plot subtitle.
    scene : str
        YAML scene file name.
    plots : bool
        If True, plot intermediate results.
    logger : `~logging.Logger` or None
        Logger for logging messages. If None, the root logger is used.
    output_dir : str or `~pathlib.Path`, optional
        Output directory for intermediate FITS files.
    """
    if logger is None:
        logger = logging.getLogger()

    # select the 2D spatial info of the 3D WCS
    wcs2d = wcs3d.sub(axes=[1, 2])

    naxis1_ifu_oversampled = Quantity(value=wcs2d.array_shape[1] * noversampling_whitelight, unit=u.pix, dtype=int)
    naxis2_ifu_oversampled = Quantity(value=wcs2d.array_shape[0] * noversampling_whitelight, unit=u.pix, dtype=int)

    bins_x_ifu_oversampled = (0.5 + np.arange(naxis1_ifu_oversampled.value + 1)) * u.pix
    bins_y_ifu_oversampled = (0.5 + np.arange(naxis2_ifu_oversampled.value + 1)) * u.pix

    crpix1_orig, crpix2_orig = wcs2d.wcs.crpix
    crpix1_oversampled = (naxis1_ifu_oversampled.value + 1) / 2
    crpix2_oversampled = (naxis2_ifu_oversampled.value + 1) / 2

    wcs2d.wcs.crpix = crpix1_oversampled, crpix2_oversampled

    if simulated_x_ifu_all is not None and simulated_y_ifu_all is not None:
        nphotons_all = len(simulated_x_ifu_all)
        if nphotons_all != len(simulated_y_ifu_all):
            raise ValueError("Length of simulated_x_ifu_all and simulated_y_ifu_all must be the same.")
    else:
        nphotons_all = 0

    if nphotons_all > 0:
        # (important: reverse X <-> Y)
        image2d_method0_ifu, xedges, yedges = np.histogram2d(
            x=(simulated_y_ifu_all.value - crpix2_orig) * noversampling_whitelight + crpix2_oversampled,
            y=(simulated_x_ifu_all.value - crpix1_orig) * noversampling_whitelight + crpix1_oversampled,
            bins=(bins_y_ifu_oversampled.value, bins_x_ifu_oversampled.value),
        )
    else:
        image2d_method0_ifu = np.zeros((naxis2_ifu_oversampled.value, naxis1_ifu_oversampled.value), dtype=int)

    wcs2d.wcs.cd /= noversampling_whitelight

    # save FITS file
    if len(prefix_intermediate_fits) > 0:
        outfile = f"{prefix_intermediate_fits}_ifu_white2D_method0_os{noversampling_whitelight:d}.fits"
        if image2d_method0_ifu.max() <= 65535:
            bitpix = 16
            hdu = fits.PrimaryHDU(image2d_method0_ifu.astype(np.uint16))
        else:
            # use float to avoid saturation problem
            logger.warning(f"The maximum value in {outfile} is greater than 65535.")
            logger.warning("Saving the image using float32 to avoid saturation.")
            bitpix = -32
            hdu = fits.PrimaryHDU(image2d_method0_ifu.astype(np.float32))
        pos0 = len(hdu.header) - 1
        hdu.header.extend(wcs_to_header_using_cd_keywords(wcs2d), update=True)
        hdu.header.update(header_keys)
        hdu.header.insert(pos0, ("COMMENT", "FITS (Flexible Image Transport System) format is defined in 'Astronomy"))
        hdu.header.insert(
            pos0 + 1, ("COMMENT", "and Astrophysics', volume 376, page 359; bibcode: 2001A&A...376..359H")
        )
        hdul = fits.HDUList([hdu])
        logger.info(f"Saving file: {outfile} (BITPIX={bitpix})")
        hdul.writeto(f"{Path(output_dir) / outfile}", overwrite=True)

    # display result
    if plots:
        fig, ax = plt.subplots(figsize=(6.4, 6.4))
        img = ax.imshow(image2d_method0_ifu, origin="lower", interpolation="None")
        ax.set_xlabel("X axis (array index)  [parallel to the slices]")
        ax.set_ylabel("Y axis (array index)  [perpendicular to the slices]")
        if instname is not None:
            title = f"{instname} "
        else:
            title = ""
        title += f"IFU image, method0 (oversampling={noversampling_whitelight})"
        if subtitle is not None:
            title += f"\n{subtitle}"
        title += f"\nscene: {scene}"
        ax.set_title(title)
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        fig.colorbar(img, cax=cax, label="Number of photons")
        plt.tight_layout()
        plt.show()



#
# Copyright 2024-2026 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#
import logging

from astropy.units import Unit
import pprint

from .raise_valueerror import raise_ValueError

pp = pprint.PrettyPrinter(indent=1, sort_dicts=False)


def set_wavelength_unit_and_range(scene_fname, scene_block, wmin, wmax, logger=None):
    """Set the wavelength unit and range for a scene block.

    Parameters
    ----------
    scene_fname : str
        YAML scene file name.
    scene_block : dict
        Dictonary storing the scene block.
    wmin : `~astropy.units.Quantity`
        Minimum wavelength covered by the detector.
    wmax : `~astropy.units.Quantity`
        Maximum wavelength covered by the detector.
    logger : logging.Logger or None, optional
        Logger for logging messages. If None, a default logger will be used.

    Returns
    -------
    wave_unit : `~astropy.units.core.Unit`
        Wavelength unit to be used in the scene block.
    wave_min : `~astropy.units.Quantity`
        Minimum wavelength to be used in the scene block.
    wave_max : `~astropy.units.Quantity`
        Maximum wavelength to be used in the scene block.

    """

    if logger is None:
        logger = logging.getLogger(__name__)

    expected_keys_in_spectrum = {"type"}

    spectrum_keys = set(scene_block["spectrum"].keys())
    if not expected_keys_in_spectrum.issubset(spectrum_keys):
        print(f"[red]ERROR while processing: {scene_fname}[/red]")
        print(f"[blue]expected keys..: {expected_keys_in_spectrum}[/blue]")
        print(f"[blue]keys found.....: {spectrum_keys}[/blue]")
        list_unexpected_keys = list(spectrum_keys.difference(expected_keys_in_spectrum))
        if len(list_unexpected_keys) > 0:
            print(f"[red]unexpected keys: {list_unexpected_keys}[/red]")
        list_missing_keys = list(expected_keys_in_spectrum.difference(spectrum_keys))
        if len(list_missing_keys) > 0:
            print(f"[red]missing keys: {list_missing_keys}[/red]")
        pp.pprint(scene_block)
        raise_ValueError(f"Invalid format in file: {scene_fname}")
    if "wave_unit" in scene_block["spectrum"]:
        wave_unit = scene_block["spectrum"]["wave_unit"]
    else:
        wave_unit = wmin.unit
        logger.debug(f"[faint]Assuming wave_unit: {wave_unit}[/faint]")
    if "wave_min" in scene_block["spectrum"]:
        wave_min = float(scene_block["spectrum"]["wave_min"])
    else:
        logger.debug(f"[faint]Assuming wave_min: null[/faint]")
        wave_min = None
    if wave_min is None:
        wave_min = wmin.to(wave_unit)
    else:
        wave_min *= Unit(wave_unit)
    if "wave_max" in scene_block["spectrum"]:
        wave_max = float(scene_block["spectrum"]["wave_max"])
    else:
        logger.debug(f"[faint]Assuming wave_max: null[/faint]")
        wave_max = None
    if wave_max is None:
        wave_max = wmax.to(wave_unit)
    else:
        wave_max = wave_max * Unit(wave_unit)

    logger.debug(f"[faint]{wave_min=}[/faint]")
    logger.debug(f"[faint]{wave_max=}[/faint]")
    logger.debug(f"[faint]{wave_unit=}[/faint]")

    return wave_unit, wave_min, wave_max



#
# Copyright 2025-2026 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#
"""Display the detected cosmic rays in the images."""

from astropy.table import Table
from matplotlib.backends.backend_pdf import PdfPages
from matplotlib.colors import ListedColormap, BoundaryNorm
import matplotlib.pyplot as plt
import numpy as np
from pathlib import Path

from numina.tools.progressbarlines import ProgressBarLines
import teareduce as tea


def display_detected_cr(
    num_images,
    image3d,
    median2d,
    median2d_corrected,
    flag_integer_dilated,
    acronym_aux,
    labels_cr,
    number_cr,
    mask_mediancr,
    list_mask_single_exposures=None,
    mask_all_singles=None,
    semiwindow=15,
    maxplots=-1,
    verify_cr=False,
    color_scale="zscale",
    xplot=None,
    yplot=None,
    boundaryfit=None,
    mm_threshold=None,
    _logger=None,
    output_dir=".",
):
    """Display the detected cosmic rays

    If list_mask_single_exposures is None, the plots correspond to the MEDIANCR
    computation. Otherwise, the plots correspond to problematic pixels (i.e.,
    cosmic-ray pixels flagged in all the single exposures).

    Note that this function can change 'mask_mediancr' if 'verify_cr' is True.
    """
    if list_mask_single_exposures is None:
        output_basename = "mediancr_identified"
        if mask_all_singles is not None:
            raise ValueError("mask_all_singles must be None when list_mask_single_exposures is None.")
    else:
        output_basename = "problematic_pixels"
        if len(list_mask_single_exposures) != num_images:
            raise ValueError("len(list_mask_single_exposures) must be equal to num_images.")
        if verify_cr:
            raise ValueError("verify_cr cannot be True when list_mask_single_exposures is provided.")
        if mask_all_singles is None:
            raise ValueError("mask_all_singles must be provided when list_mask_single_exposures is provided.")

    # Determine the number of rows and columns for the plot,
    # considering that we want to plot also 3 additional images:
    # the median2d, the mask and the median2d_corrected
    num_plot_max = num_images
    if num_images == 3:
        nrows, ncols = 2, 3
        figsize = (10, 6)
    elif num_images in [4, 5]:
        nrows, ncols = 2, 4
        figsize = (13, 6)
    elif num_images == 6:
        nrows, ncols = 3, 3
        figsize = (13, 3)
    elif num_images in [7, 8, 9]:
        nrows, ncols = 3, 4
        figsize = (13, 9)
    else:
        _logger.warning("only the first 9 images will be plotted")
        nrows, ncols = 3, 4
        figsize = (13, 9)
        num_plot_max = 9
    if list_mask_single_exposures is None:
        # plots for mediancr
        pdf_any4 = PdfPages(Path(output_dir) / f"{output_basename}_any4.pdf")
        pdf_only3 = PdfPages(Path(output_dir) / f"{output_basename}_only3.pdf")
        pdf_only2 = PdfPages(Path(output_dir) / f"{output_basename}_only2.pdf")
        pdf_other = PdfPages(Path(output_dir) / f"{output_basename}_other.pdf")
        num_any4 = 0
        num_only3 = 0
        num_only2 = 0
        num_other = 0
        cr_table_any4 = Table(names=("CR_number", "X_pixel", "Y_pixel", "Mask_value"), dtype=(int, int, int, int))
        cr_table_only3 = Table(names=("CR_number", "X_pixel", "Y_pixel", "Mask_value"), dtype=(int, int, int, int))
        cr_table_only2 = Table(names=("CR_number", "X_pixel", "Y_pixel", "Mask_value"), dtype=(int, int, int, int))
        cr_table_other = Table(names=("CR_number", "X_pixel", "Y_pixel", "Mask_value"), dtype=(int, int, int, int))
        cr_table_filename_any4 = f"{output_basename}_any4.csv"
        cr_table_filename_only3 = f"{output_basename}_only3.csv"
        cr_table_filename_only2 = f"{output_basename}_only2.csv"
        cr_table_filename_other = f"{output_basename}_other.csv"
    else:
        # plots for problematic pixels
        pdf = PdfPages(Path(output_dir) / f"{output_basename}.pdf")
        cr_table = Table(names=("CR_number", "X_pixel", "Y_pixel", "Mask_value"), dtype=(int, int, int, int))
        cr_table_filename = f"{output_basename}.csv"
    maxplots_eff = maxplots
    if verify_cr:
        # In verify_cr mode, we plot all the cosmic rays
        maxplots_eff = number_cr
    elif maxplots_eff < 0:
        if number_cr > 1000:
            maxplots_eff = 1000
            _logger.warning(f"limiting to {maxplots_eff} plots (out of {number_cr} CRs detected)")
        else:
            maxplots_eff = number_cr
    _logger.info(f"generating {maxplots_eff} plots...")
    xlabel = "X pixel (from 1 to NAXIS1)"
    ylabel = "Y pixel (from 1 to NAXIS2)"
    naxis2, naxis1 = median2d.shape
    total = min(number_cr, maxplots_eff)
    pbar = ProgressBarLines(total, _logger)
    for i in range(total):
        # Locate the cosmic ray pixels
        ijloc = np.argwhere(labels_cr == i + 1)
        ic = int(np.mean(ijloc[:, 0]) + 0.5)
        jc = int(np.mean(ijloc[:, 1]) + 0.5)
        i1 = ic - semiwindow
        if i1 < 0:
            i1 = 0
        i2 = ic + semiwindow
        if i2 >= naxis2:
            i2 = naxis2 - 1
        j1 = jc - semiwindow
        if j1 < 0:
            j1 = 0
        j2 = jc + semiwindow
        if j2 >= naxis1:
            j2 = naxis1 - 1
        # Generate the plots
        fig, axarr = plt.subplots(nrows=nrows, ncols=ncols, figsize=figsize)
        axarr = axarr.flatten()
        # Important: use interpolation=None instead of interpolation='None' to avoid
        # having blurred images when opening the PDF file with macos Preview
        cmap = "viridis"
        cblabel = "Number of counts"
        if color_scale == "zscale":
            vmin, vmax = tea.zscale(median2d[i1 : (i2 + 1), j1 : (j2 + 1)])
        else:
            vmin = np.min(median2d[i1 : (i2 + 1), j1 : (j2 + 1)])
            vmax = np.max(median2d[i1 : (i2 + 1), j1 : (j2 + 1)])
        for k in range(num_plot_max):
            ax = axarr[k]
            title = title = f"image#{k+1}/{num_images}"
            tea.imshow(
                fig,
                ax,
                image3d[k][i1 : (i2 + 1), j1 : (j2 + 1)],
                vmin=vmin,
                vmax=vmax,
                extent=[j1 + 0.5, j2 + 1.5, i1 + 0.5, i2 + 1.5],
                xlabel=xlabel,
                ylabel=ylabel,
                title=title,
                cmap=cmap,
                cblabel=cblabel,
                interpolation=None,
            )
        if list_mask_single_exposures is None:
            # Plots for mediancr
            for k in range(3):
                ax = axarr[k + num_plot_max]
                cmap = "viridis"
                if k == 0:
                    image2d = median2d
                    title = "median"
                    norm = None
                elif k == 1:
                    image2d = flag_integer_dilated
                    title = "flag_integer_dilated"
                    # cmap = 'plasma'
                    color_list = ["black", "grey", "blue", "red", "yellow"]
                    cmap = ListedColormap(color_list)
                    bounds = np.arange(-0.5, len(color_list), 1)  # integer limits: -0.5, 0.5, 1.5, ...
                    norm = BoundaryNorm(bounds, cmap.N)
                    cblabel = None
                elif k == 2:
                    image2d = median2d_corrected
                    title = "median corrected"
                    norm = None
                else:
                    raise ValueError(f"Unexpected {k=}")
                if k in [0, 2]:
                    if color_scale == "zscale":
                        vmin_, vmax_ = tea.zscale(image2d[i1 : (i2 + 1), j1 : (j2 + 1)])
                    else:
                        vmin_ = np.min(image2d[i1 : (i2 + 1), j1 : (j2 + 1)])
                        vmax_ = np.max(image2d[i1 : (i2 + 1), j1 : (j2 + 1)])
                else:
                    vmin_, vmax_ = None, None
                img, cax, cbar = tea.imshow(
                    fig,
                    ax,
                    image2d[i1 : (i2 + 1), j1 : (j2 + 1)],
                    vmin=vmin_,
                    vmax=vmax_,
                    extent=[j1 + 0.5, j2 + 1.5, i1 + 0.5, i2 + 1.5],
                    xlabel=xlabel,
                    ylabel=ylabel,
                    title=title,
                    cmap=cmap,
                    norm=norm,
                    cblabel=cblabel,
                    interpolation=None,
                )
                if k == 1:
                    cbar.set_ticks([0, 1, 2, 3, 4])
                    cbar.ax.yaxis.set_tick_params(length=0)
                    cbar.set_ticklabels(
                        ["0: no CR", "1: dilation", "2: mm", f"3: {acronym_aux}", f"4: mm & {acronym_aux}"]
                    )
                    # Overlay the grid of pixels
                    for idum in range(i1, i2 + 1):
                        ax.hlines(y=idum + 0.5, xmin=j1 + 0.5, xmax=j2 + 1.5, colors="w", lw=0.5, alpha=0.3)
                    for jdum in range(j1, j2 + 1):
                        ax.vlines(x=jdum + 0.5, ymin=i1 + 0.5, ymax=i2 + 1.5, colors="w", lw=0.5, alpha=0.3)
            nplot_missing = nrows * ncols - num_plot_max - 3
            if nplot_missing > 0:
                for k in range(nplot_missing):
                    ax = axarr[-k - 1]
                    ax.axis("off")
            fig.suptitle(f"CR#{i+1}/{number_cr}")
        else:
            # Plot the problematic pixels
            ijloc = np.argwhere(labels_cr == i + 1)
            xproblematic = ijloc[:, 1] + 1  # FITS criterium
            yproblematic = ijloc[:, 0] + 1  # FITS criter
            for k in range(num_plot_max):
                ax = axarr[k + num_plot_max]
                title = title = f"mask#{k+1}/{num_images}"
                image2d = list_mask_single_exposures[k].astype(int)
                tea.imshow(
                    fig,
                    ax,
                    image2d[i1 : (i2 + 1), j1 : (j2 + 1)],
                    vmin=0,
                    vmax=1,
                    extent=[j1 + 0.5, j2 + 1.5, i1 + 0.5, i2 + 1.5],
                    xlabel=xlabel,
                    ylabel=ylabel,
                    title=title,
                    cmap="plasma",
                    cblabel="flag",
                    interpolation=None,
                )
                for xdum, ydum in zip(xproblematic, yproblematic):
                    ax.plot([xdum - 0.5, xdum + 0.5], [ydum - 0.5, ydum + 0.5], "r-")
                    ax.plot([xdum - 0.5, xdum + 0.5], [ydum + 0.5, ydum - 0.5], "r-")
            fig.suptitle(f"Problematic pixels, region #{i+1}/{number_cr}")

        plt.tight_layout()

        if verify_cr:
            plt.show(block=False)
            print("-" * 50)
        different_flag_values = []
        for idum in range(i1, i2 + 1):
            for jdum in range(j1, j2 + 1):
                if labels_cr[idum, jdum] == i + 1:
                    different_flag_values.append(flag_integer_dilated[idum, jdum])
        different_flag_values = set(different_flag_values)
        for idum in range(i1, i2 + 1):
            for jdum in range(j1, j2 + 1):
                if labels_cr[idum, jdum] == i + 1:
                    if list_mask_single_exposures is None:
                        if 4 in different_flag_values:
                            cr_table_any4.add_row((i + 1, jdum + 1, idum + 1, flag_integer_dilated[idum, jdum]))
                        elif 3 in different_flag_values and 2 not in different_flag_values:
                            cr_table_only3.add_row((i + 1, jdum + 1, idum + 1, flag_integer_dilated[idum, jdum]))
                        elif 2 in different_flag_values and 3 not in different_flag_values:
                            cr_table_only2.add_row((i + 1, jdum + 1, idum + 1, flag_integer_dilated[idum, jdum]))
                        else:
                            cr_table_other.add_row((i + 1, jdum + 1, idum + 1, flag_integer_dilated[idum, jdum]))
                    else:
                        cr_table.add_row((i + 1, jdum + 1, idum + 1, flag_integer_dilated[idum, jdum]))
                    if verify_cr:
                        print(f"pixel (x,y) = ({jdum+1}, {idum+1}):  mask = {flag_integer_dilated[idum, jdum]}")
        if verify_cr:
            accept_cr = input(f"Accept this CR detection #{i+1}: " "[Y]es (default) | (n)o | (a)ll | (s)top)? ")
            if accept_cr.lower() == "n":
                _logger.info("removing cosmic ray detection #%d from the mask\n", i + 1)
                mask_mediancr[labels_cr == i + 1] = False
            elif accept_cr.lower() == "a":
                verify_cr = False
                _logger.info("accepting all remaining cosmic ray detections\n")
                _logger.info("generating remaining plots without user interaction...")
            elif accept_cr.lower() == "s":
                _logger.info("stopping the program execution")
                raise SystemExit(0)
            else:
                _logger.info("keeping cosmic ray detection #%d in the mask", i + 1)
        if list_mask_single_exposures is None:
            # plot for mediancr
            if 4 in different_flag_values:
                pdf_any4.savefig(fig, bbox_inches="tight")
                num_any4 += 1
            elif 3 in different_flag_values and 2 not in different_flag_values:
                pdf_only3.savefig(fig, bbox_inches="tight")
                num_only3 += 1
            elif 2 in different_flag_values and 3 not in different_flag_values:
                pdf_only2.savefig(fig, bbox_inches="tight")
                num_only2 += 1
            else:
                pdf_other.savefig(fig, bbox_inches="tight")
                num_other += 1
        else:
            # plot for problematic pixels
            pdf.savefig(fig, bbox_inches="tight")
        plt.close(fig)

        if list_mask_single_exposures is None:
            # plot pixels in the M.M. diagnostic diagram
            fig, ax = plt.subplots(figsize=figsize)
            if boundaryfit is not None:
                xplot_boundary = np.linspace(np.min(xplot), np.max(xplot), 1000)
                yplot_boundary = boundaryfit(xplot_boundary)
            else:
                raise ValueError("boundaryfit must be provided for mediancr plots")
            ax.plot(xplot_boundary, yplot_boundary, color="C1", linestyle="-", label="detection boundary")
            if mm_threshold is not None:
                ax.axhline(y=mm_threshold, color="C0", linestyle=":", label="mm_threshold")
            legend4 = True
            legend3 = True
            legend2 = True
            legend0 = True
            for idum in range(i1, i2 + 1):
                for jdum in range(j1, j2 + 1):
                    if labels_cr[idum, jdum] == i + 1:
                        if flag_integer_dilated[idum, jdum] == 4:
                            symbol = "mo"
                            if legend4:
                                label = f"mm & {acronym_aux}"
                                legend4 = False
                            else:
                                label = None
                        elif flag_integer_dilated[idum, jdum] == 3:
                            symbol = "rx"
                            if legend3:
                                label = f"{acronym_aux}"
                                legend3 = False
                            else:
                                label = None
                        elif flag_integer_dilated[idum, jdum] == 2:
                            symbol = "b+"
                            if legend2:
                                label = "mm"
                                legend2 = False
                            else:
                                label = None
                        else:
                            symbol = "ko"
                            if legend0:
                                label = "dilation"
                                legend0 = False
                            else:
                                label = None
                        xval = xplot.reshape((naxis2, naxis1))[idum, jdum]
                        yval = yplot.reshape((naxis2, naxis1))[idum, jdum]
                        ax.plot(xval, yval, symbol, label=label)
                        ax.text(xval, yval, f"({jdum+1},{idum+1})", fontsize=6, color="gray")
            _, ymax = ax.get_ylim()
            ax.set_ylim(0.0, ymax)
            ax.set_xlabel(r"min2d $-$ bias  [ADU]")
            ax.set_ylabel(r"median2d $-$ min2d  [ADU]")
            ax.legend()
            fig.suptitle(f"CR#{i+1}/{number_cr}")
            plt.tight_layout()
            if 4 in different_flag_values:
                pdf_any4.savefig(fig, bbox_inches="tight")
            elif 3 in different_flag_values and 2 not in different_flag_values:
                pdf_only3.savefig(fig, bbox_inches="tight")
            elif 2 in different_flag_values and 3 not in different_flag_values:
                pdf_only2.savefig(fig, bbox_inches="tight")
            else:
                pdf_other.savefig(fig, bbox_inches="tight")
            plt.close(fig)
        pbar.update()

    _logger.info("plot generation complete")
    if list_mask_single_exposures is None:
        # plots for mediancr
        num_max = np.max([num_any4, num_only3, num_only2, num_other])
        snum_max = str(num_max)
        lmax = len(snum_max)
        pdf_any4.close()
        _logger.info(f"saving {num_any4:>{lmax}} CRs in {output_basename}_any4.pdf file")
        pdf_only3.close()
        _logger.info(f"saving {num_only3:>{lmax}} CRs in {output_basename}_only3.pdf file")
        pdf_only2.close()
        _logger.info(f"saving {num_only2:>{lmax}} CRs in {output_basename}_only2.pdf file")
        pdf_other.close()
        _logger.info(f"saving {num_other:>{lmax}} CRs in {output_basename}_other.pdf file")
        cr_table_any4.write(Path(output_dir) / cr_table_filename_any4, format="csv", overwrite=True)
        _logger.info(f"table of {num_any4} CRs in {output_basename}_any4.csv:\n{cr_table_any4}")
        _logger.info("table saved")
        cr_table_only3.write(Path(output_dir) / cr_table_filename_only3, format="csv", overwrite=True)
        _logger.info(f"table of {num_only3} CRs in {output_basename}_only3.csv:\n{cr_table_only3}")
        _logger.info("table saved")
        cr_table_only2.write(Path(output_dir) / cr_table_filename_only2, format="csv", overwrite=True)
        _logger.info(f"table of {num_only2} CRs in {output_basename}_only2.csv:\n{cr_table_only2}")
        _logger.info("table saved")
        cr_table_other.write(Path(output_dir) / cr_table_filename_other, format="csv", overwrite=True)
        _logger.info(f"table of {num_other} CRs in {output_basename}_other.csv:\n{cr_table_other}")
        _logger.info("table saved")
    else:
        # plots for problematic pixels
        pdf.close()
        _logger.info(f"saving {output_basename}.pdf file")
        cr_table.write(Path(output_dir) / cr_table_filename, format="csv", overwrite=True)
        _logger.info("\n%s", cr_table)
        _logger.info("table of identified cosmic rays saved to %s", cr_table_filename)
